import ast
from copy import deepcopy

import numpy as np
import pandas as pd
from conftest import ASSETS
from scipy.stats import linregress

from qmllib.kernels import (
    get_atomic_local_gradient_kernel,
    get_atomic_local_kernel,
    get_gp_kernel,
    get_symmetric_gp_kernel,
)
from qmllib.representations import generate_fchl19
from qmllib.solvers import cho_solve, svd_solve

np.set_printoptions(linewidth=999, edgeitems=10, suppress=True)


TRAINING = 7
TEST = 5
VALID = 3

ELEMENTS = [1, 6, 7, 8]

CUT_DISTANCE = 8.0

DF_TRAIN = pd.read_csv(ASSETS / "force_train.csv", delimiter=";").head(TRAINING)
DF_VALID = pd.read_csv(ASSETS / "force_valid.csv", delimiter=";").head(VALID)
DF_TEST = pd.read_csv(ASSETS / "force_test.csv", delimiter=";").head(TEST)

SIGMA = 21.2

LAMBDA_ENERGY = 1e-6
LAMBDA_FORCE = 1e-6

np.random.seed(666)


def get_reps(df):

    x = []
    f = []
    e = []
    disp_x = []
    q = []

    max_atoms = 27

    for i in range(len(df)):

        coordinates = np.array(ast.literal_eval(df["coordinates"][i]))
        nuclear_charges = np.array(ast.literal_eval(df["nuclear_charges"][i]), dtype=np.int32)
        # UNUSED atomtypes = df["atomtypes"][i]

        force = np.array(ast.literal_eval(df["forces"][i]))
        force *= -1

        energy = float(df["atomization_energy"][i])

        (x1, dx1) = generate_fchl19(nuclear_charges, coordinates, gradients=True, pad=max_atoms)

        x.append(x1)
        f.append(force)
        e.append(energy)

        disp_x.append(dx1)
        q.append(nuclear_charges)

    e = np.array(e)
    # e -= np.mean(e)# - 10 #

    # print(f)

    # f = np.array(f)
    # f *= -1
    x = np.array(x)

    return x, f, e, np.array(disp_x), q


def test_fchl_acsf_operator():

    print("Representations ...")
    X, F, E, dX, Q = get_reps(DF_TRAIN)
    Xs, Fs, Es, dXs, Qs = get_reps(DF_TEST)
    Xv, Fv, Ev, dXv, Qv = get_reps(DF_VALID)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)
    Fv = np.concatenate(Fv)

    Y = np.concatenate((E, F.flatten()))

    print("Kernels ...")
    Kte = get_atomic_local_kernel(X, X, Q, Q, SIGMA)
    Kse = get_atomic_local_kernel(X, Xs, Q, Qs, SIGMA)
    Kve = get_atomic_local_kernel(X, Xv, Q, Qv, SIGMA)

    Kt = get_atomic_local_gradient_kernel(X, X, dX, Q, Q, SIGMA)
    Ks = get_atomic_local_gradient_kernel(X, Xs, dXs, Q, Qs, SIGMA)
    Kv = get_atomic_local_gradient_kernel(X, Xv, dXv, Q, Qv, SIGMA)

    C = np.concatenate((Kte, Kt))

    print("Alphas operator ...")
    alpha = svd_solve(C, Y, rcond=1e-11)
    # alpha = qrlq_solve(C, Y)

    eYt = np.dot(Kte, alpha)
    eYs = np.dot(Kse, alpha)
    eYv = np.dot(Kve, alpha)

    fYt = np.dot(Kt, alpha)
    fYs = np.dot(Ks, alpha)
    fYv = np.dot(Kv, alpha)

    print(
        "==============================================================================================="
    )
    print(
        "====  OPERATOR, FORCE + ENERGY  ==============================================================="
    )
    print(
        "==============================================================================================="
    )

    slope, intercept, r_value, p_value, std_err = linregress(E, eYt)
    print(
        "TRAINING ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(E - eYt)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(F.flatten(), fYt.flatten())
    print(
        "TRAINING FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(F.flatten() - fYt.flatten())), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Es.flatten(), eYs.flatten())
    print(
        "TEST     ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Es - eYs)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Fs.flatten(), fYs.flatten())
    print(
        "TEST     FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Fs.flatten() - fYs.flatten())), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Ev.flatten(), eYv.flatten())
    print(
        "VALID    ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Ev - eYv)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Fv.flatten(), fYv.flatten())
    print(
        "VALID    FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Fv.flatten() - fYv.flatten())), slope, intercept, r_value)
    )


def test_fchl_acsf_gaussian_process():

    print("Representations ...")
    X, F, E, dX, Q = get_reps(DF_TRAIN)
    Xs, Fs, Es, dXs, Qs = get_reps(DF_TEST)
    Xv, Fv, Ev, dXv, Qv = get_reps(DF_VALID)

    F = np.concatenate(F)
    Fs = np.concatenate(Fs)
    Fv = np.concatenate(Fv)

    print("Kernels ...")
    Kt_gp = get_symmetric_gp_kernel(X, dX, Q, SIGMA)
    Ks_gp = get_gp_kernel(X, Xs, dX, dXs, Q, Qs, SIGMA)
    Kv_gp = get_gp_kernel(X, Xv, dX, dXv, Q, Qv, SIGMA)

    print("Alphas GP ...")
    C = deepcopy(Kt_gp)

    for i in range(TRAINING):
        C[i, i] += LAMBDA_ENERGY

    for i in range(TRAINING, C.shape[0]):
        C[i, i] += LAMBDA_FORCE

    Y = np.concatenate((E, F.flatten()))

    alpha = cho_solve(C, Y)

    Yt = np.dot(Kt_gp, alpha)
    Ys = np.dot(Ks_gp, alpha)
    Yv = np.dot(Kv_gp, alpha)

    eYt = Yt[:TRAINING]
    eYs = Ys[:TEST]
    eYv = Yv[:VALID]

    fYt = Yt[TRAINING:]
    fYs = Ys[TEST:]
    fYv = Yv[VALID:]

    print(
        "==============================================================================================="
    )
    print(
        "====  GAUSSIAN PROCESS, FORCE + ENERGY  ======================================================="
    )
    print(
        "==============================================================================================="
    )

    slope, intercept, r_value, p_value, std_err = linregress(E, eYt)
    print(
        "TRAINING ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(E - eYt)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(F.flatten(), fYt.flatten())
    print(
        "TRAINING FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(F.flatten() - fYt.flatten())), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Es.flatten(), eYs.flatten())
    print(
        "TEST     ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Es - eYs)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Fs.flatten(), fYs.flatten())
    print(
        "TEST     FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Fs.flatten() - fYs.flatten())), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Ev.flatten(), eYv.flatten())
    print(
        "VALID    ENERGY   MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Ev - eYv)), slope, intercept, r_value)
    )

    slope, intercept, r_value, p_value, std_err = linregress(Fv.flatten(), fYv.flatten())
    print(
        "VALID    FORCE    MAE = %10.4f  slope = %10.4f  intercept = %10.4f  r^2 = %9.6f"
        % (np.mean(np.abs(Fv.flatten() - fYv.flatten())), slope, intercept, r_value)
    )


if __name__ == "__main__":

    test_fchl_acsf_operator()
    test_fchl_acsf_gaussian_process()



from typing import List, Union

import numpy as np
from numpy import float64, ndarray

from .fkernels import (
    fgaussian_kernel,
    fgaussian_kernel_symmetric,
    fget_local_kernels_gaussian,
    fget_local_kernels_laplacian,
    fkpca,
    flaplacian_kernel,
    flaplacian_kernel_symmetric,
    flinear_kernel,
    fmatern_kernel_l2,
    fsargan_kernel,
    fwasserstein_kernel,
)


def wasserstein_kernel(A: ndarray, B: ndarray, sigma: float, p: int = 1, q: int = 1) -> ndarray:
    """Calculates the Wasserstein kernel matrix K, where :math:`K_{ij}`:

    :math:`K_{ij} = \\exp \\big( -\\frac{(W_p(A_i, B_i))^q}{\\sigma} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Wasserstein kernel matrix - shape (N, M)
    :rtype: numpy array
    """

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    fwasserstein_kernel(A.T, na, B.T, nb, K, sigma, p, q)

    return K


def laplacian_kernel(A: ndarray, B: ndarray, sigma: float) -> ndarray:
    """Calculates the Laplacian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - B_j\\|_1}{\\sigma} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Laplacian kernel matrix - shape (N, M)
    :rtype: numpy array
    """

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    flaplacian_kernel(A.T, na, B.T, nb, K, sigma)

    return K


def laplacian_kernel_symmetric(A: ndarray, sigma: float) -> ndarray:
    """Calculates the symmetric Laplacian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - A_j\\|_1}{\\sigma} \\big)`

    Where :math:`A_{i}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Laplacian kernel matrix - shape (N, N)
    :rtype: numpy array
    """

    na = A.shape[0]

    K = np.empty((na, na), order="F")

    # Note: Transposed for Fortran
    flaplacian_kernel_symmetric(A.T, na, K, sigma)

    return K


def gaussian_kernel(A: ndarray, B: ndarray, sigma: float) -> ndarray:
    """Calculates the Gaussian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - B_j\\|_2^2}{2\\sigma^2} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Gaussian kernel matrix - shape (N, M)
    :rtype: numpy array
    """

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    fgaussian_kernel(A.T, na, B.T, nb, K, sigma)

    return K


def gaussian_kernel_symmetric(A: ndarray, sigma: float) -> ndarray:
    """Calculates the symmetric Gaussian kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\|A_i - A_j\\|_2^2}{2\\sigma^2} \\big)`

    Where :math:`A_{i}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Gaussian kernel matrix - shape (N, N)
    :rtype: numpy array
    """

    na = A.shape[0]

    K = np.empty((na, na), order="F")

    # Note: Transposed for Fortran
    fgaussian_kernel_symmetric(A.T, na, K, sigma)

    return K


def linear_kernel(A: ndarray, B: ndarray) -> ndarray:
    """Calculates the linear kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = A_i \\cdot B_j`

    VWhere :math:`A_{i}` and :math:`B_{j}` are  representation vectors.

    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array

    :return: The Gaussian kernel matrix - shape (N, M)
    :rtype: numpy array
    """

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    flinear_kernel(A.T, na, B.T, nb, K)

    return K


def sargan_kernel(
    A: ndarray,
    B: ndarray,
    sigma: Union[float, float64],
    gammas: Union[ndarray, List[Union[int, float]], List[int]],
) -> ndarray:
    """Calculates the Sargan kernel matrix K, where :math:`K_{ij}`:

        :math:`K_{ij} = \\exp \\big( -\\frac{\\| A_i - B_j \\|_1)}{\\sigma} \\big) \\big(1 + \\sum_{k} \\frac{\\gamma_{k} \\| A_i - B_j \\|_1^k}{\\sigma^k} \\big)`

    Where :math:`A_{i}` and :math:`B_{j}` are representation vectors.
    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float
    :param gammas: 1D array of parameters in the kernel matrix.
    :type gammas: numpy array

    :return: The Sargan kernel matrix - shape (N, M).
    :rtype: numpy array
    """

    ng = len(gammas)

    if ng == 0:
        return laplacian_kernel(A, B, sigma)

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    fsargan_kernel(A.T, na, B.T, nb, K, sigma, gammas, ng)

    return K


def matern_kernel(
    A: ndarray, B: ndarray, sigma: float, order: int = 0, metric: str = "l1"
) -> ndarray:
    """Calculates the Matern kernel matrix K, where :math:`K_{ij}`:

        for order = 0:
            :math:`K_{ij} = \\exp\\big( -\\frac{d}{\\sigma} \\big)`
        for order = 1:
            :math:`K_{ij} = \\exp\\big( -\\frac{\\sqrt{3} d}{\\sigma} \\big) \\big(1 + \\frac{\\sqrt{3} d}{\\sigma} \\big)`
        for order = 2:
            :math:`K_{ij} = \\exp\\big( -\\frac{\\sqrt{5} d}{d} \\big) \\big( 1 + \\frac{\\sqrt{5} d}{\\sigma} + \\frac{5 d^2}{3\\sigma^2} \\big)`

    Where :math:`A_i` and :math:`B_j` are representation vectors, and d is a distance measure.

    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of representations - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of representations - shape (M, representation size).
    :type B: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float
    :param order: The order of the polynomial (0, 1, 2)
    :type order: integer
    :param metric: The distance metric ('l1', 'l2')
    :type metric: string

    :return: The Matern kernel matrix - shape (N, M)
    :rtype: numpy array
    """

    if metric == "l1":

        if order == 0:
            gammas = []

        elif order == 1:
            gammas = [1]
            sigma /= np.sqrt(3)

        elif order == 2:
            gammas = [1, 1 / 3.0]
            sigma /= np.sqrt(5)

        else:
            raise ValueError(f"Order '{order}' not implemented in Matern Kernel")

        return sargan_kernel(A, B, sigma, gammas)

    elif metric == "l2":
        pass

    else:
        raise ValueError(f"Unknown distance metric {metric} in Matern kernel")

    na = A.shape[0]
    nb = B.shape[0]

    K = np.empty((na, nb), order="F")

    # Note: Transposed for Fortran
    fmatern_kernel_l2(A.T, na, B.T, nb, K, sigma, order)

    return K


def get_local_kernels_gaussian(
    A: ndarray, B: ndarray, na: ndarray, nb: ndarray, sigmas: List[float]
) -> ndarray:
    """Calculates the Gaussian kernel matrix K, for a local representation where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{a \\in i} \\sum_{b \\in j} \\exp \\big( -\\frac{\\|A_a - B_b\\|_2^2}{2\\sigma^2} \\big)`

    Where :math:`A_{a}` and :math:`B_{b}` are representation vectors.

    Note that the input array is one big 2D array with all atoms concatenated along the same axis.
    Further more a series of kernels is produced (since calculating the distance matrix is expensive
    but getting the resulting kernels elements for several sigmas is not.)

    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of descriptors - shape (total atoms A, representation size).
    :type A: numpy array
    :param B: 2D array of descriptors - shape (total atoms B, representation size).
    :type B: numpy array
    :param na: 1D array containing numbers of atoms in each compound.
    :type na: numpy array
    :param nb: 1D array containing numbers of atoms in each compound.
    :type nb: numpy array
    :param sigma: The value of sigma in the kernel matrix.
    :type sigma: float

    :return: The Gaussian kernel matrix - shape (nsigmas, N, M)
    :rtype: numpy array
    """

    if np.sum(na) != A.shape[0]:
        raise ValueError("Error in A input")
    if np.sum(nb) != B.shape[0]:
        raise ValueError("Error in B input")

    if A.shape[1] != B.shape[1]:
        raise ValueError("Error in representation sizes")

    nma = len(na)
    nmb = len(nb)

    sigmas = np.asarray(sigmas)
    nsigmas = len(sigmas)

    return fget_local_kernels_gaussian(A.T, B.T, na, nb, sigmas, nma, nmb, nsigmas)


def get_local_kernels_laplacian(
    A: ndarray, B: ndarray, na: ndarray, nb: ndarray, sigmas: List[float]
) -> ndarray:
    """Calculates the Local Laplacian kernel matrix K, for a local representation where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{a \\in i} \\sum_{b \\in j} \\exp \\big( -\\frac{\\|A_a - B_b\\|_1}{\\sigma} \\big)`

    Where :math:`A_{a}` and :math:`B_{b}` are representation vectors.

    Note that the input array is one big 2D array with all atoms concatenated along the same axis.
    Further more a series of kernels is produced (since calculating the distance matrix is expensive
    but getting the resulting kernels elements for several sigmas is not.)

    K is calculated using an OpenMP parallel Fortran routine.

    :param A: 2D array of descriptors - shape (N, representation size).
    :type A: numpy array
    :param B: 2D array of descriptors - shape (M, representation size).
    :type B: numpy array
    :param na: 1D array containing numbers of atoms in each compound.
    :type na: numpy array
    :param nb: 1D array containing numbers of atoms in each compound.
    :type nb: numpy array
    :param sigmas: List of the sigmas.
    :type sigmas: list

    :return: The Laplacian kernel matrix - shape (nsigmas, N, M)
    :rtype: numpy array
    """

    if np.sum(na) != A.shape[0]:
        raise ValueError("Error in A input")
    if np.sum(nb) != B.shape[0]:
        raise ValueError("Error in B input")

    if A.shape[1] != B.shape[1]:
        raise ValueError("Error in representation sizes")

    nma = len(na)
    nmb = len(nb)

    sigmas = np.asarray(sigmas)
    nsigmas = len(sigmas)

    return fget_local_kernels_laplacian(A.T, B.T, na, nb, sigmas, nma, nmb, nsigmas)


def kpca(K: ndarray, n: int = 2, centering: bool = True) -> ndarray:
    """Calculates `n` first principal components for the kernel :math:`K`.

    The PCA is calculated using an OpenMP parallel Fortran routine.

    A square, symmetric kernel matrix is required. Centering of the kernel matrix
    is enabled by default, although this isn't a strict requirement.

    :param K: 2D kernel matrix
    :type K: numpy array
    :param n: Number of kernel PCAs to return (default=2)
    :type n: integer
    :param centering: Whether to center the kernel matrix (default=True)
    :type centering: bool

    :return: array containing the principal components
    :rtype: numpy array
    """

    if K.shape[0] != K.shape[1]:
        raise ValueError("Square matrix required for Kernel PCA.")
    if not np.allclose(K, K.T, atol=1e-8):
        raise ValueError("Symmetric matrix required for Kernel PCA.")
    if not n <= K.shape[0]:
        raise ValueError("Requested more principal components than matrix size.")

    size = K.shape[0]
    pca = fkpca(K, size, centering)

    return pca[:n]



from copy import deepcopy

import numpy as np
from conftest import ASSETS

from qmllib.solvers import bkf_invert, bkf_solve, cho_invert, cho_solve


def test_cho_solve():

    A_ref = np.loadtxt(ASSETS / "K_local_gaussian.txt")
    y_ref = np.loadtxt(ASSETS / "y_cho_solve.txt")

    A = deepcopy(A_ref)
    y = deepcopy(y_ref)
    x_qml = cho_solve(A, y)

    # Check arrays are unchanged
    assert np.allclose(y, y_ref)
    assert np.allclose(A, A_ref)

    A = deepcopy(A_ref)
    x_scipy = np.linalg.solve(A, y)

    # Check for correct solution
    assert np.allclose(x_qml, x_scipy)


def test_cho_invert():

    A_ref = np.loadtxt(ASSETS / "K_local_gaussian.txt")

    A = deepcopy(A_ref)
    Ai_qml = cho_invert(A)

    # Check A is unchanged
    assert np.allclose(A, A_ref)

    A = deepcopy(A_ref)
    one = np.eye(A.shape[0])

    # Check that it is a true inverse
    assert np.allclose(np.matmul(A, Ai_qml), one, atol=1e-7)


def test_bkf_invert():

    A_ref = np.loadtxt(ASSETS / "K_local_gaussian.txt")

    A = deepcopy(A_ref)
    Ai_qml = bkf_invert(A)

    # Check A is unchanged
    assert np.allclose(A, A_ref)

    A = deepcopy(A_ref)
    one = np.eye(A.shape[0])

    np.set_printoptions(linewidth=20000)
    assert np.allclose(np.matmul(A, Ai_qml), one, atol=1e-7)


def test_bkf_solve():

    A_ref = np.loadtxt(ASSETS / "K_local_gaussian.txt")
    y_ref = np.loadtxt(ASSETS / "y_cho_solve.txt")

    A = deepcopy(A_ref)
    y = deepcopy(y_ref)
    x_qml = bkf_solve(A, y)

    # Check arrays are unchanged
    assert np.allclose(y, y_ref)
    assert np.allclose(A, A_ref)

    A = deepcopy(A_ref)
    y = deepcopy(y_ref)
    x_scipy = np.linalg.solve(A, y)

    # Check for correct solution
    assert np.allclose(x_qml, x_scipy)


if __name__ == "__main__":
    test_cho_solve()
    test_cho_invert()
    test_bkf_invert()
    test_bkf_solve()



import numpy as np
from conftest import ASSETS

from qmllib.representations import generate_slatm, get_slatm_mbtypes
from qmllib.utils.xyz_format import read_xyz


def test_slatm_global_representation():

    files = [
        ASSETS / "qm7/0001.xyz",
        ASSETS / "qm7/0002.xyz",
        ASSETS / "qm7/0003.xyz",
        ASSETS / "qm7/0004.xyz",
        ASSETS / "qm7/0005.xyz",
        ASSETS / "qm7/0006.xyz",
        ASSETS / "qm7/0007.xyz",
        ASSETS / "qm7/0008.xyz",
        ASSETS / "qm7/0009.xyz",
        ASSETS / "qm7/0010.xyz",
    ]

    mols = []
    for xyz_file in files:
        coordinates, atoms = read_xyz(xyz_file)
        mols.append((coordinates, atoms))

    charges = [atoms for _, atoms in mols]

    mbtypes = get_slatm_mbtypes(charges)
    print("mbtypes:", mbtypes)

    representations = []
    for coord, atoms in mols:
        slatm_vector = generate_slatm(atoms, coord, mbtypes)
        representations.append(slatm_vector)

    X_qml = np.array([rep for rep in representations])
    X_ref = np.loadtxt(ASSETS / "slatm_global_representation.txt")

    assert np.allclose(X_qml, X_ref), "Error in SLATM generation"


def test_slatm_local_representation():

    files = [
        ASSETS / "qm7/0001.xyz",
        ASSETS / "qm7/0002.xyz",
        ASSETS / "qm7/0003.xyz",
        ASSETS / "qm7/0004.xyz",
        ASSETS / "qm7/0005.xyz",
        ASSETS / "qm7/0006.xyz",
        ASSETS / "qm7/0007.xyz",
        ASSETS / "qm7/0008.xyz",
        ASSETS / "qm7/0009.xyz",
        ASSETS / "qm7/0010.xyz",
    ]

    mols = []
    for xyz_file in files:
        coordinates, atoms = read_xyz(xyz_file)
        mols.append((coordinates, atoms))

    charges = [atoms for _, atoms in mols]
    mbtypes = get_slatm_mbtypes(charges)

    local_representations = []
    for _, mol in enumerate(mols):

        coord, atoms = mol
        slatm_vector = generate_slatm(atoms, coord, mbtypes, local=True)

        local_representations.append(slatm_vector)

    # Spread the structures into atom vectors
    X_qml = []
    for representation in local_representations:
        for rep in representation:
            X_qml.append(rep)

    X_qml = np.asarray(X_qml)
    X_ref = np.loadtxt(ASSETS / "slatm_local_representation.txt")

    assert np.allclose(X_qml, X_ref), "Error in SLATM generation"


if __name__ == "__main__":

    test_slatm_global_representation()
    test_slatm_local_representation()



"""
This file contains tests for the atom centred symmetry function module.
"""

from copy import deepcopy

import numpy as np
from conftest import ASSETS

from qmllib.representations import generate_fchl19
from qmllib.utils.xyz_format import read_xyz

np.set_printoptions(linewidth=666, edgeitems=10)
REP_PARAMS = dict()
REP_PARAMS["elements"] = [1, 6, 7]


def get_acsf_numgrad(coordinates, nuclear_charges, dx=1e-5):

    natoms = len(coordinates)
    true_coords = deepcopy(coordinates)

    true_rep = generate_fchl19(nuclear_charges, coordinates, gradients=False, **REP_PARAMS)

    gradient = np.zeros((3, natoms, true_rep.shape[0], true_rep.shape[1]))

    for n, coord in enumerate(true_coords):
        for xyz, x in enumerate(coord):

            temp_coords = deepcopy(true_coords)
            temp_coords[n, xyz] = x + 2.0 * dx

            (rep, grad) = generate_fchl19(
                nuclear_charges, temp_coords, gradients=True, **REP_PARAMS
            )
            gradient[xyz, n] -= rep

            temp_coords[n, xyz] = x + dx
            (rep, grad) = generate_fchl19(
                nuclear_charges, temp_coords, gradients=True, **REP_PARAMS
            )
            gradient[xyz, n] += 8.0 * rep

            temp_coords[n, xyz] = x - dx
            (rep, grad) = generate_fchl19(
                nuclear_charges, temp_coords, gradients=True, **REP_PARAMS
            )
            gradient[xyz, n] -= 8.0 * rep

            temp_coords[n, xyz] = x - 2.0 * dx
            (rep, grad) = generate_fchl19(
                nuclear_charges, temp_coords, gradients=True, **REP_PARAMS
            )
            gradient[xyz, n] += rep

    gradient /= 12 * dx

    gradient = np.swapaxes(gradient, 0, 1)
    gradient = np.swapaxes(gradient, 2, 0)
    gradient = np.swapaxes(gradient, 3, 1)

    return gradient


def test_fchl_acsf():

    coordinates, nuclear_charges = read_xyz(ASSETS / "qm7/0101.xyz")

    (repa, anal_grad) = generate_fchl19(nuclear_charges, coordinates, gradients=True, **REP_PARAMS)

    repb = generate_fchl19(nuclear_charges, coordinates, gradients=False, **REP_PARAMS)

    assert np.allclose(repa, repb), "Error in FCHL-ACSF representation implementation"

    num_grad = get_acsf_numgrad(coordinates, nuclear_charges)

    assert np.allclose(anal_grad, num_grad), "Error in FCHL-ACSF gradient implementation"



import numpy as np

from qmllib.kernels.distance import l2_distance, manhattan_distance, p_distance


def test_manhattan():

    nfeatures = 5
    n1 = 7
    n2 = 9

    v1 = np.random.random((n1, nfeatures))
    v2 = np.random.random((n2, nfeatures))

    D = manhattan_distance(v1, v2)

    Dtest = np.zeros((n1, n2))

    for i in range(n1):
        for j in range(n2):
            for k in range(nfeatures):
                Dtest[i, j] += abs(v1[i, k] - v2[j, k])

    assert np.allclose(D, Dtest), "Error in manhattan distance"


def test_l2():

    nfeatures = 5
    n1 = 7
    n2 = 9

    v1 = np.random.random((n1, nfeatures))
    v2 = np.random.random((n2, nfeatures))

    D = l2_distance(v1, v2)

    Dtest = np.zeros((n1, n2))

    for i in range(n1):
        for j in range(n2):
            for k in range(nfeatures):
                Dtest[i, j] += (v1[i, k] - v2[j, k]) ** 2

    np.sqrt(Dtest, out=Dtest)

    assert np.allclose(D, Dtest), "Error in l2 distance"


def test_p():

    nfeatures = 5
    n1 = 7
    n2 = 9

    v1 = np.random.random((n1, nfeatures))
    v2 = np.random.random((n2, nfeatures))

    D = p_distance(v1, v2, 3)

    Dtest = np.zeros((n1, n2))

    for i in range(n1):
        for j in range(n2):
            for k in range(nfeatures):
                Dtest[i, j] += abs(v1[i, k] - v2[j, k]) ** 3

    Dtest = Dtest ** (1.0 / 3)

    assert np.allclose(D, Dtest), "Error in p-distance"

    Dfloat = p_distance(v1, v2, 3.0)
    assert np.allclose(D, Dfloat), "Floatingpoint Error in p-distance"


if __name__ == "__main__":
    test_manhattan()
    test_l2()
    test_p()



import numpy as np
from conftest import ASSETS

from qmllib.representations import (
    generate_bob,
    generate_coulomb_matrix,
    generate_coulomb_matrix_atomic,
    generate_coulomb_matrix_eigenvalue,
)
from qmllib.representations.bob import get_asize
from qmllib.utils.xyz_format import read_xyz


def _get_molecules():
    files = [
        ASSETS / "qm7/0101.xyz",
        ASSETS / "qm7/0102.xyz",
        ASSETS / "qm7/0103.xyz",
        ASSETS / "qm7/0104.xyz",
        ASSETS / "qm7/0105.xyz",
        ASSETS / "qm7/0106.xyz",
        ASSETS / "qm7/0107.xyz",
        ASSETS / "qm7/0108.xyz",
        ASSETS / "qm7/0109.xyz",
        ASSETS / "qm7/0110.xyz",
    ]

    mols = []
    for filename in files:
        coordinates, atoms = read_xyz(filename)
        mols.append((coordinates, atoms))

    return mols

    # size = max(atoms.size for _, atoms in mols) + 1

    # asize = get_asize([atoms for atoms in mols], 1)

    # coulomb_matrix(mols, size, path)
    # atomic_coulomb_matrix(mols, size, path)
    # eigenvalue_coulomb_matrix(mols, size, path)
    # bob(mols, size, asize, path)


def test_coulomb_matrix_rownorm():

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    # Generate coulomb matrix representation, sorted by row-norm
    representations = []
    for coordinates, nuclear_charges in mols:
        representation = generate_coulomb_matrix(
            nuclear_charges, coordinates, size=size, sorting="row-norm"
        )
        representations.append(representation)

    X_test = np.asarray([rep for rep in representations])

    print(X_test.shape)

    X_ref = np.loadtxt(ASSETS / "coulomb_matrix_representation_row-norm_sorted.txt")
    assert np.allclose(X_test, X_ref), "Error in coulomb matrix representation"


def test_coulomb_matrix_unsorted():

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    # Generate coulomb matrix representation, unsorted, using the Compound class
    representations = []
    for coordinates, nuclear_charges in mols:
        representation = generate_coulomb_matrix(
            nuclear_charges, coordinates, size=size, sorting="unsorted"
        )
        representations.append(representation)

    X_test = np.asarray([rep for rep in representations])

    print(X_test.shape)

    X_ref = np.loadtxt(ASSETS / "coulomb_matrix_representation_unsorted.txt")
    assert np.allclose(X_test, X_ref), "Error in coulomb matrix representation"


def test_atomic_coulomb_matrix_distance():

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    # Generate coulomb matrix representation, sorted by distance
    representations = []
    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_atomic(nuclear_charges, coord, size=size, sorting="distance")
        representations.append(rep)

    X_test = np.concatenate([rep for rep in representations])
    X_ref = np.loadtxt(ASSETS / "atomic_coulomb_matrix_representation_distance_sorted.txt")
    assert np.allclose(X_test, X_ref), "Error in atomic coulomb matrix representation"
    # Compare to old implementation (before 'indices' keyword)
    X_ref = np.loadtxt(
        ASSETS / "atomic_coulomb_matrix_representation_distance_sorted_no_indices.txt"
    )
    assert np.allclose(X_test, X_ref), "Error in atomic coulomb matrix representation"


def test_atomic_coulomb_matrix_rownorm():

    # Generate coulomb matrix representation, sorted by row-norm

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    representations = []
    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_atomic(nuclear_charges, coord, size=size, sorting="row-norm")
        representations.append(rep)

    X_test = np.concatenate(representations)
    X_ref = np.loadtxt(ASSETS / "atomic_coulomb_matrix_representation_row-norm_sorted.txt")
    assert np.allclose(X_test, X_ref), "Error in atomic coulomb matrix representation"


def test_atomic_coulomb_matrix_distance_softcut():

    # Generate coulomb matrix representation, sorted by distance, with soft cutoffs

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    representations = []

    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_atomic(
            nuclear_charges,
            coord,
            size=size,
            sorting="distance",
            central_cutoff=4.0,
            central_decay=0.5,
            interaction_cutoff=5.0,
            interaction_decay=1.0,
        )
        representations.append(rep)

    X_test = np.concatenate([rep for rep in representations])
    X_ref = np.loadtxt(
        ASSETS / "atomic_coulomb_matrix_representation_distance_sorted_with_cutoff.txt"
    )
    assert np.allclose(X_test, X_ref), "Error in atomic coulomb matrix representation"


def test_atomic_coulomb_matrix_rownorm_cut():

    # Generate coulomb matrix representation, sorted by row-norm, with soft cutoffs

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    representations = []

    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_atomic(
            nuclear_charges,
            coord,
            size=size,
            sorting="row-norm",
            central_cutoff=4.0,
            central_decay=0.5,
            interaction_cutoff=5.0,
            interaction_decay=1.0,
        )
        representations.append(rep)

    X_test = np.concatenate(representations)
    X_ref = np.loadtxt(
        ASSETS / "atomic_coulomb_matrix_representation_row-norm_sorted_with_cutoff.txt"
    )
    assert np.allclose(X_test, X_ref), "Error in atomic coulomb matrix representation"


def test_atomic_coulomb_matrix_twoatom_distance():

    # Generate only two atoms in the coulomb matrix representation, sorted by distance

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_atomic(nuclear_charges, coord, size=size, sorting="distance")
        representation_subset = rep[1:3]
        rep = generate_coulomb_matrix_atomic(
            nuclear_charges, coord, size=size, sorting="distance", indices=[1, 2]
        )
        for i in range(2):
            for j in range(153):
                diff = representation_subset[i, j] - rep[i, j]
                if abs(diff) > 1e-9:
                    print(i, j, diff, representation_subset[i, j], rep[i, j])

        assert np.allclose(
            representation_subset, rep
        ), "Error in atomic coulomb matrix representation"


def test_atomic_coulomb_matrix_twoatom_rownorm():

    # Generate only two atoms in the coulomb matrix representation, sorted by row-norm

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    for coord, nuclear_charges in mols:

        rep = generate_coulomb_matrix_atomic(nuclear_charges, coord, size=size, sorting="row-norm")
        representation_subset = rep[1:3]
        rep = generate_coulomb_matrix_atomic(
            nuclear_charges, coord, size=size, sorting="row-norm", indices=[1, 2]
        )
        for i in range(2):
            for j in range(153):
                diff = representation_subset[i, j] - rep[i, j]
                if abs(diff) > 1e-9:
                    print(i, j, diff, representation_subset[i, j], rep[i, j])
        assert np.allclose(
            representation_subset, rep
        ), "Error in atomic coulomb matrix representation"


def test_eigenvalue_coulomb_matrix():

    # Generate coulomb matrix representation, sorted by row-norm

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    representations = []

    for coord, nuclear_charges in mols:
        rep = generate_coulomb_matrix_eigenvalue(nuclear_charges, coord, size=size)
        representations.append(rep)

    X_test = np.asarray(representations)
    X_ref = np.loadtxt(ASSETS / "eigenvalue_coulomb_matrix_representation.txt")
    assert np.allclose(X_test, X_ref), "Error in eigenvalue coulomb matrix representation"


def test_bob():

    mols = _get_molecules()
    size = max(atoms.size for _, atoms in mols) + 1

    # example asize={"O": 3, "C": 7, "N": 3, "H": 16, "S": 1},
    asize = get_asize([atoms for _, atoms in mols], 1)

    atomtypes = []
    for _, atoms in mols:
        atomtypes.extend(atoms)
    atomtypes = np.unique(atomtypes)

    print(size)
    print(atomtypes)
    print(asize)

    representations = []

    for coord, nuclear_charges in mols:
        rep = generate_bob(nuclear_charges, coord, atomtypes, size=size, asize=asize)
        representations.append(rep)

    X_test = np.asarray(representations)
    X_ref = np.loadtxt(ASSETS / "bob_representation.txt")
    assert np.allclose(X_test, X_ref), "Error in bag of bonds representation"



from typing import Optional

import numpy as np
from numpy import ndarray

from .fsolvers import (
    fbkf_invert,
    fbkf_solve,
    fcho_invert,
    fcho_solve,
    fcond,
    fcond_ge,
    fqrlq_solve,
    fsvd_solve,
)


def cho_invert(A: ndarray) -> ndarray:
    """Returns the inverse of a positive definite matrix, using a Cholesky decomposition
    via calls to LAPACK dpotrf and dpotri in the F2PY module.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array

    :return: The inverse matrix
    :rtype: numpy array
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected square matrix")

    matrix = np.asfortranarray(A)

    fcho_invert(matrix)

    # Matrix to store the inverse
    i_lower = np.tril_indices_from(A)

    # Copy lower triangle to upper
    matrix.T[i_lower] = matrix[i_lower]

    return matrix


def cho_solve(A: ndarray, y: ndarray, l2reg: float = 0.0, destructive: bool = False) -> ndarray:
    """Solves the equation

        :math:`A x = y`

    for x using a Cholesky decomposition  via calls to LAPACK dpotrf and dpotrs in the F2PY module. Preserves the input matrix A.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array
    :param y: Vector (right-hand side of the equation).
    :type y: numpy array
    :param l2reg: Small number to add to the diagonal as L2-regularization when solving.
    :type l2reg: float
    :param destructive: Whether to preserve the lower triangle after solving(=False) or destroy it, which is faster(=True).
    :type destructive: bool


    :return: The solution vector.
    :rtype: numpy array
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected square matrix")

    if len(y.shape) != 1 or y.shape[0] != A.shape[1]:
        raise ValueError("expected matrix and vector of same stride size")

    n = A.shape[0]

    # Backup diagonal before Cholesky-decomposition
    A_diag = A[np.diag_indices_from(A)]

    for i in range(len(y)):

        A[i, i] += l2reg

    x = np.zeros(n)
    fcho_solve(A, y, x)

    # Reset diagonal after Cholesky-decomposition
    A[np.diag_indices_from(A)] = A_diag

    if destructive is False:

        # Copy lower triangle to upper
        i_lower = np.tril_indices_from(A)
        A.T[i_lower] = A[i_lower]

    return x


def bkf_invert(A: ndarray) -> ndarray:
    """Returns the inverse of a positive definite matrix, using a Bausch-Kauffman decomposition
    via calls to LAPACK dpotrf and dpotri in the F2PY module.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array

    :return: The inverse matrix
    :rtype: numpy array
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected square matrix")

    matrix = np.asfortranarray(A)

    fbkf_invert(matrix)

    # Matrix to store the inverse
    i_lower = np.tril_indices_from(A)

    # Copy lower triangle to upper
    matrix.T[i_lower] = matrix[i_lower]

    return matrix


def bkf_solve(A: ndarray, y: ndarray) -> ndarray:
    """Solves the equation

        :math:`A x = y`

    for x using a  Bausch-Kauffma  decomposition  via calls to LAPACK  in the F2PY module. Preserves the input matrix A.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array
    :param y: Vector (right-hand side of the equation).
    :type y: numpy array

    :return: The solution vector.
    :rtype: numpy array
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected square matrix")

    if len(y.shape) != 1 or y.shape[0] != A.shape[1]:
        raise ValueError("expected matrix and vector of same stride size")

    n = A.shape[0]

    # Backup diagonal before Cholesky-decomposition
    A_diag = A[np.diag_indices_from(A)]

    x = np.zeros(n)
    fbkf_solve(A, y, x)

    # Reset diagonal after Cholesky-decomposition
    A[np.diag_indices_from(A)] = A_diag

    # Copy lower triangle to upper
    i_lower = np.tril_indices_from(A)
    A.T[i_lower] = A[i_lower]

    return x


def svd_solve(A: ndarray, y: ndarray, rcond: Optional[float] = None) -> ndarray:
    """Solves the equation

        :math:`A x = y`

    for x using a singular-value decomposition (SVD) via calls to
    LAPACK DGELSD in the F2PY module. Preserves the input matrix A.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array
    :param y: Vector (right-hand side of the equation).
    :type y: numpy array
    :param rcond: Optional parameater for lowest singular-value
    :type rcond: float

    :return: The solution vector.
    :rtype: numpy array
    """

    if len(y.shape) != 1 or y.shape[0] != A.shape[0]:
        raise ValueError("expected matrix and vector of same stride size")

    if rcond is None:
        rcond = 0.0

    x_dim = A.shape[1]
    A = np.asarray(A, order="F")
    x = fsvd_solve(A, y, x_dim, rcond)

    return x


def qrlq_solve(A, y):
    """Solves the equation

        :math:`A x = y`

    for x using a QR or LQ decomposition (depending on matrix dimensions)
    via calls to LAPACK DGELSD in the F2PY module. Preserves the input matrix A.

    :param A: Matrix (symmetric and positive definite, left-hand side).
    :type A: numpy array
    :param y: Vector (right-hand side of the equation).
    :type y: numpy array

    :return: The solution vector.
    :rtype: numpy array
    """

    if len(y.shape) != 1 or y.shape[0] != A.shape[0]:
        raise ValueError("expected matrix and vector of same stride size")

    x_dim = A.shape[1]
    A = np.asarray(A, order="F")
    x = fqrlq_solve(A, y, x_dim)

    return x


def condition_number(A, method="cholesky"):
    """Returns the condition number for the square matrix A.

    Two different methods are implemented:
    Cholesky (requires a positive-definite matrix), but barely any additional memory overhead.
    LU: Does not require a positive definite matrix, but requires additional memory.
    """

    if len(A.shape) != 2 or A.shape[0] != A.shape[1]:
        raise ValueError("expected square matrix")

    if method.lower() == "cholesky":

        if not np.allclose(A, A.T):
            raise ValueError("Can't use a Cholesky-decomposition for a non-symmetric matrix.")

        cond = fcond(A)

        return cond

    elif method.lower() == "lu":

        cond = fcond_ge(A)

        return cond



from typing import List, Union

import numpy as np
from numpy import ndarray

from qmllib.utils.environment_manipulation import (
    mkl_get_num_threads,
    mkl_reset_num_threads,
    mkl_set_num_threads,
)

from .fgradient_kernels import (
    fatomic_local_gradient_kernel,
    fatomic_local_kernel,
    fgaussian_process_kernel,
    fgdml_kernel,
    fglobal_kernel,
    flocal_gradient_kernel,
    flocal_kernel,
    flocal_kernels,
    fsymmetric_gaussian_process_kernel,
    fsymmetric_gdml_kernel,
    fsymmetric_local_kernel,
    fsymmetric_local_kernels,
)


def get_global_kernel(
    X1: ndarray, X2: ndarray, Q1: List[List[int]], Q2: List[List[int]], SIGMA: float
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{I\\in i} \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    The kernel is the normal Gaussian kernel with the local decomposition of atomic environments.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("Error: List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    K = fglobal_kernel(X1, X2, Q1_input, Q2_input, N1, N2, len(N1), len(N2), SIGMA)

    return K


def get_local_kernels(
    X1: ndarray, X2: ndarray, Q1: List[List[int]], Q2: List[List[int]], SIGMAS: List[float]
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{I\\in i} \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    The kernel is the normal Gaussian kernel with the local decomposition of atomic environments.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("Error: List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("Error: List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    sigmas_input = np.array(SIGMAS, dtype=np.float64)
    nsigmas = len(SIGMAS)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    K = flocal_kernels(X1, X2, Q1_input, Q2_input, N1, N2, len(N1), len(N2), sigmas_input, nsigmas)

    return K


def get_local_kernel(
    X1: ndarray,
    X2: ndarray,
    Q1: List[Union[ndarray, List[int]]],
    Q2: List[Union[ndarray, List[int]]],
    SIGMA: float,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{I\\in i} \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    The kernel is the normal Gaussian kernel with the local decomposition of atomic environments.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    K = flocal_kernel(X1, X2, Q1_input, Q2_input, N1, N2, len(N1), len(N2), SIGMA)

    return K


def get_local_symmetric_kernels(X1: ndarray, Q1: List[List[int]], SIGMAS: List[float]) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{I\\in i} \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    The kernel is the normal Gaussian kernel with the local decomposition of atomic environments.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("Error: List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    nsigmas = len(SIGMAS)
    K = fsymmetric_local_kernels(X1, Q1_input, N1, len(N1), SIGMAS, nsigmas)

    return K


def get_local_symmetric_kernel(
    X1: ndarray, Q1: List[Union[ndarray, List[int]]], SIGMA: float
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\sum_{I\\in i} \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    The kernel is the normal Gaussian kernel with the local decomposition of atomic environments.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("Error: List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    K = fsymmetric_local_kernel(X1, Q1_input, N1, len(N1), SIGMA)

    return K


def get_atomic_local_kernel(
    X1: ndarray,
    X2: ndarray,
    Q1: List[Union[ndarray, List[int]]],
    Q2: List[Union[ndarray, List[int]]],
    SIGMA: float,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{Ij} = \\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.

    This means that the kernel matrix consists of the local decomposition of atomic environments
    in a basis of kernel function placed on single atomic environments.

    Thus the dimensions in number of molecules times total number of atoms.

    For instance atom-centered symmetry functions could be used here.
    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    K = fatomic_local_kernel(
        X1, X2, Q1_input, Q2_input, N1, N2, len(N1), len(N2), np.sum(N1), SIGMA
    )

    return K


def get_atomic_local_gradient_kernel(
    X1: ndarray,
    X2: ndarray,
    dX2: ndarray,
    Q1: List[Union[ndarray, List[int]]],
    Q2: List[Union[ndarray, List[int]]],
    SIGMA: float,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{Ij} = \\frac{\\part}{\\part x}\\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.
    For instance atom-centered symmetry functions could be used here.

    The kernel has the dimensions number of nuclear kernel gradients times total number of atoms.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param dX2: Array of representation derivatives - shape=(N2, rep_size, 3, rep_size, max_atoms).
    :type dX2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = fatomic_local_gradient_kernel(
        X1,
        X2,
        dX2,
        Q1_input,
        Q2_input,
        N1,
        N2,
        len(N1),
        len(N2),
        np.sum(N1),
        np.sum(N2) * 3,
        SIGMA,
    )

    # Reset MKL_NUM_THREADS back to its original value
    if original_mkl_threads is not None:
        mkl_set_num_threads(original_mkl_threads)
    else:
        mkl_reset_num_threads()

    return K


def get_local_gradient_kernel(
    X1: ndarray, X2: ndarray, dX2: ndarray, Q1: List[List[int]], Q2: List[List[int]], SIGMA: float
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{ij} = \\frac{\\part}{\\part x}\\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.
    For instance atom-centered symmetry functions could be used here.

    The kernel has the dimensions number of nuclear kernel gradients times total number of molecules.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param dX2: Array of representation derivatives - shape=(N2, rep_size, 3, rep_size, max_atoms).
    :type dX2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = flocal_gradient_kernel(
        X1, X2, dX2, Q1_input, Q2_input, N1, N2, len(N1), len(N2), np.sum(N2) * 3, SIGMA
    )

    # Reset MKL_NUM_THREADS back to its original value
    mkl_set_num_threads(original_mkl_threads)

    return K


def get_gdml_kernel(
    X1: ndarray,
    X2: ndarray,
    dX1: ndarray,
    dX2: ndarray,
    Q1: List[List[int]],
    Q2: List[List[int]],
    SIGMA: float,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{Ij} = \\frac{\\part^2}{\\part x_i\\part x_j}\\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.
    For instance atom-centered symmetry functions could be used here.

    This Hessian-kernel corresponds to the "gradient-domain machine learning" (GDML) approach.
    This means that the surface is only trained on its derivatives.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param dX1: Array of representation derivatives - shape=(N1, rep_size, 3, rep_size, max_atoms).
    :type dX1: numpy array

    :param dX2: Array of representation derivatives - shape=(N2, rep_size, 3, rep_size, max_atoms).
    :type dX2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = fgdml_kernel(
        X1,
        X2,
        dX1,
        dX2,
        Q1_input,
        Q2_input,
        N1,
        N2,
        len(N1),
        len(N2),
        np.sum(N1),
        np.sum(N2),
        SIGMA,
    )

    # Reset MKL_NUM_THREADS back to its original value
    mkl_set_num_threads(original_mkl_threads)

    return K


def get_symmetric_gdml_kernel(
    X1: ndarray, dX1: ndarray, Q1: List[List[int]], SIGMA: float
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

        :math:`K_{Ij} = \\frac{\\part^2}{\\part x_i\\part x_j}\\sum_{J\\in j}\\exp \\big( -\\frac{\\|X_I - X_J\\|_2^2}{2\\sigma^2} \\big)`

    Where :math: X_{I}` and :math:`X_{J}` are representation vectors of the atomic environments.
    For instance atom-centered symmetry functions could be used here.

    This symmetric Hessian-kernel corresponds to the "gradient-domain machine learning" (GDML) approach.
    This means that the surface is only trained on its derivatives.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array

    :param dX1: Array of representation derivatives - shape=(N1, rep_size, 3, rep_size, max_atoms).
    :type dX1: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = fsymmetric_gdml_kernel(X1, dX1, Q1_input, N1, len(N1), np.sum(N1), SIGMA)

    # Reset MKL_NUM_THREADS back to its original value
    mkl_set_num_threads(original_mkl_threads)

    return K


def get_gp_kernel(
    X1: ndarray,
    X2: ndarray,
    dX1: ndarray,
    dX2: ndarray,
    Q1: List[Union[ndarray, List[int]]],
    Q2: List[Union[ndarray, List[int]]],
    SIGMA: float,
) -> ndarray:
    """Calculates the Gaussian kernel matrix K with the local decomposition where :math:`K_{ij}`:

    This kernel corresponds to a Gaussian process regression (GPR) approach.
    The kernel has four blocks, consisting of the 0'th, 1st and 2nd derivatives.

    The size is (the number of gradients plus the number of molecules) squared.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array
    :param X2: Array of representations - shape=(N2, rep_size, max_atoms).
    :type X2: numpy array

    :param dX1: Array of representation derivatives - shape=(N1, rep_size, 3, rep_size, max_atoms).
    :type dX1: numpy array

    :param dX2: Array of representation derivatives - shape=(N2, rep_size, 3, rep_size, max_atoms).
    :type dX2: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)
    N2 = np.array([len(Q) for Q in Q2], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")
    if not (N2.shape[0] == X2.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)
    Q2_input = np.zeros((max(N2), X2.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    for i, q in enumerate(Q2):
        Q2_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = fgaussian_process_kernel(
        X1,
        X2,
        dX1,
        dX2,
        Q1_input,
        Q2_input,
        N1,
        N2,
        len(N1),
        len(N2),
        np.sum(N1),
        np.sum(N2),
        SIGMA,
    )

    # Reset MKL_NUM_THREADS back to its original value
    mkl_set_num_threads(original_mkl_threads)

    return K


def get_symmetric_gp_kernel(
    X1: ndarray, dX1: ndarray, Q1: List[Union[ndarray, List[int]]], SIGMA: float
) -> ndarray:
    """
    This symmetric kernel corresponds to a Gaussian process regression (GPR) approach.
    The kernel has four blocks, consisting of the 0'th, 1st and 2nd derivatives.

    The size is (the number of gradients plus the number of molecules) squared.

    K is calculated analytically using an OpenMP parallel Fortran routine.

    :param X1: Array of representations - shape=(N1, rep_size, max_atoms).
    :type X1: numpy array

    :param dX1: Array of representation derivatives - shape=(N1, rep_size, 3, rep_size, max_atoms).
    :type dX1: numpy array

    :param Q1: List of lists containing the nuclear charges for each molecule.
    :type Q1: list
    :param Q2: List of lists containing the nuclear charges for each molecule.
    :type Q2: list

    :param SIGMA: Gaussian kernel width.
    :type SIGMA: float

    :return: 2D matrix of kernel elements shape=(N1, N2),
    :rtype: numpy array
    """

    N1 = np.array([len(Q) for Q in Q1], dtype=np.int32)

    if not (N1.shape[0] == X1.shape[0]):
        raise ValueError("List of charges does not match shape of representations")

    Q1_input = np.zeros((max(N1), X1.shape[0]), dtype=np.int32)

    for i, q in enumerate(Q1):
        Q1_input[: len(q), i] = q

    # This kernel must run with MKL_NUM_THREADS=1
    original_mkl_threads = mkl_get_num_threads()
    mkl_set_num_threads(1)

    K = fsymmetric_gaussian_process_kernel(X1, dX1, Q1_input, N1, len(N1), np.sum(N1), SIGMA)

    # Reset MKL_NUM_THREADS back to its original value
    mkl_set_num_threads(original_mkl_threads)

    return K



from .fchl_electric_field_kernels import *  # noqa:F403
from .fchl_force_kernels import *  # noqa:F403
from .fchl_representations import *  # noqa:F403
from .fchl_scalar_kernels import *  # noqa:F403



import numpy as np

from qmllib.utils.alchemy import get_alchemy

from .fchl_kernel_functions import get_kernel_parameters
from .ffchl_module import (
    fget_atomic_local_gradient_5point_kernels_fchl,
    fget_atomic_local_gradient_kernels_fchl,
    fget_force_alphas_fchl,
    fget_gaussian_process_kernels_fchl,
    fget_local_gradient_kernels_fchl,
    fget_local_hessian_kernels_fchl,
    fget_local_symmetric_hessian_kernels_fchl,
)


def get_gaussian_process_kernels(
    A,
    B,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = B.shape[4]
    neighbors_max = B.shape[6]

    if (
        not B.shape[1] == 3
        or not B.shape[2] == 2
        or not B.shape[5] == 5
        or not A.shape[2] == 5
        or not A.shape[1] == atoms_max
        or not B.shape[3] == atoms_max
        or not A.shape[3] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors2 = np.zeros((nm2, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    naq2 = np.sum(N2) * 3

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    return fget_gaussian_process_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        naq2,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )


def get_local_gradient_kernels(
    A,
    B,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = B.shape[4]

    neighbors_max = B.shape[6]

    if (
        not B.shape[1] == 3
        or not B.shape[2] == 2
        or not B.shape[5] == 5
        or not A.shape[2] == 5
        or not A.shape[1] == atoms_max
        or not B.shape[3] == atoms_max
        or not A.shape[3] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors2 = np.zeros((nm2, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    naq2 = np.sum(N2) * 3

    return fget_local_gradient_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        naq2,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )


def get_local_hessian_kernels(
    A,
    B,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = A.shape[4]

    neighbors_max = A.shape[6]

    if (
        not A.shape[1] == 3
        or not A.shape[2] == 2
        or not A.shape[5] == 5
        or not B.shape[1] == 3
        or not B.shape[2] == 2
        or not B.shape[5] == 5
        or not A.shape[3] == atoms_max
        or not B.shape[3] == atoms_max
        or not B.shape[6] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)
    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, 3, 2, atoms_max, atoms_max), dtype=np.int32)
    neighbors2 = np.zeros((nm2, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm1):
        ni = N1[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(A[m, xyz, pm, i, :ni]):
                        neighbors1[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    naq1 = np.sum(N1) * 3
    naq2 = np.sum(N2) * 3

    # print naq1, naq2, nsigmas
    return fget_local_hessian_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        naq1,
        naq2,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )


def get_local_symmetric_hessian_kernels(
    A,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]

    atoms_max = A.shape[4]

    # Unused neighbors_max = A.shape[6]

    if (
        not A.shape[1] == 3
        or not A.shape[2] == 2
        or not A.shape[5] == 5
        or not A.shape[3] == atoms_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm1):
        ni = N1[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(A[m, xyz, pm, i, :ni]):
                        neighbors1[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    naq1 = np.sum(N1) * 3

    return fget_local_symmetric_hessian_kernels_fchl(
        A,
        verbose,
        N1,
        neighbors1,
        nm1,
        naq1,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )


def get_force_alphas(
    A,
    B,
    F,
    verbose=False,
    energy=None,
    dx=0.005,
    regularization=1e-7,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):
    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = B.shape[4]

    neighbors_max = B.shape[6]

    if (
        not B.shape[1] == 3
        or not B.shape[2] == 2
        or not B.shape[5] == 5
        or not A.shape[2] == 5
        or not A.shape[1] == atoms_max
        or not B.shape[3] == atoms_max
        or not A.shape[3] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors2 = np.zeros((nm2, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    na1 = np.sum(N1)
    # UNUSED naq2 = np.sum(N2) * 3

    E = np.zeros((nm1))
    if energy is not None:
        E = energy

    return fget_force_alphas_fchl(
        A,
        B,
        verbose,
        F,
        E,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        na1,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
        regularization,
    )


def get_atomic_local_gradient_kernels(
    A,
    B,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = B.shape[4]

    neighbors_max = B.shape[6]

    if (
        not B.shape[1] == 3
        or not B.shape[2] == 2
        or not B.shape[5] == 5
        or not A.shape[2] == 5
        or not A.shape[1] == atoms_max
        or not B.shape[3] == atoms_max
        or not A.shape[3] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors2 = np.zeros((nm2, 3, 2, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(2):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    na1 = np.sum(N1)
    naq2 = np.sum(N2) * 3

    return fget_atomic_local_gradient_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        na1,
        naq2,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )


def get_atomic_local_gradient_5point_kernels(
    A,
    B,
    verbose=False,
    dx=0.005,
    two_body_scaling=np.sqrt(8),
    three_body_scaling=1.6,
    two_body_width=0.2,
    three_body_width=np.pi,
    two_body_power=4.0,
    three_body_power=2.0,
    cut_start=1.0,
    cut_distance=5.0,
    fourier_order=1,
    alchemy="periodic-table",
    alchemy_period_width=1.6,
    alchemy_group_width=1.6,
    kernel="gaussian",
    kernel_args=None,
):

    nm1 = A.shape[0]
    nm2 = B.shape[0]

    atoms_max = B.shape[4]

    neighbors_max = B.shape[6]

    if (
        not B.shape[1] == 3
        or not B.shape[2] == 5
        or not B.shape[5] == 5
        or not A.shape[2] == 5
        or not A.shape[1] == atoms_max
        or not B.shape[3] == atoms_max
        or not A.shape[3] == neighbors_max
    ):
        raise ValueError("Representations have wrong shape")

    N1 = np.zeros((nm1), dtype=np.int32)

    for a in range(nm1):
        N1[a] = len(np.where(A[a, :, 1, 0] > 0.0001)[0])

    neighbors1 = np.zeros((nm1, atoms_max), dtype=np.int32)

    for a, representation in enumerate(A):
        ni = N1[a]
        for i, x in enumerate(representation[:ni]):
            neighbors1[a, i] = len(np.where(x[0] < cut_distance)[0])

    N2 = np.zeros((nm2), dtype=np.int32)

    for a in range(nm2):
        N2[a] = len(np.where(B[a, 0, 0, 0, :, 1, 0] > 0.0001)[0])

    neighbors2 = np.zeros((nm2, 3, 5, atoms_max, atoms_max), dtype=np.int32)

    for m in range(nm2):
        ni = N2[m]
        for xyz in range(3):
            for pm in range(5):
                for i in range(ni):
                    for a, x in enumerate(B[m, xyz, pm, i, :ni]):
                        neighbors2[m, xyz, pm, i, a] = len(np.where(x[0] < cut_distance)[0])

    doalchemy, pd = get_alchemy(
        alchemy, emax=100, r_width=alchemy_group_width, c_width=alchemy_period_width
    )

    kernel_idx, kernel_parameters, nsigmas = get_kernel_parameters(kernel, kernel_args)

    na1 = np.sum(N1)
    naq2 = np.sum(N2) * 3

    return fget_atomic_local_gradient_5point_kernels_fchl(
        A,
        B,
        verbose,
        N1,
        N2,
        neighbors1,
        neighbors2,
        nm1,
        nm2,
        na1,
        naq2,
        nsigmas,
        three_body_width,
        two_body_width,
        cut_start,
        cut_distance,
        fourier_order,
        pd,
        two_body_scaling,
        three_body_scaling,
        doalchemy,
        two_body_power,
        three_body_power,
        dx,
        kernel_idx,
        kernel_parameters,
    )



import numpy as np
from conftest import ASSETS, get_energies
from scipy.stats import wasserstein_distance
from sklearn.decomposition import KernelPCA

from qmllib.kernels import (
    gaussian_kernel,
    gaussian_kernel_symmetric,
    kpca,
    laplacian_kernel,
    laplacian_kernel_symmetric,
    linear_kernel,
    matern_kernel,
    sargan_kernel,
    wasserstein_kernel,
)
from qmllib.representations import generate_bob
from qmllib.utils.xyz_format import read_xyz


def test_laplacian_kernel():

    np.random.seed(666)

    n_train = 25
    n_test = 20

    # List of dummy representations
    X = np.random.rand(n_train, 1000)
    Xs = np.random.rand(n_test, 1000)

    sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):
            Ktest[i, j] = np.exp(np.sum(np.abs(X[i] - Xs[j])) / (-1.0 * sigma))

    K = laplacian_kernel(X, Xs, sigma)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in Laplacian kernel"

    Ksymm = laplacian_kernel(X, X, sigma)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in Laplacian kernel"

    Ksymm2 = laplacian_kernel_symmetric(X, sigma)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm2), "Error in Laplacian kernel"


def test_gaussian_kernel():

    np.random.seed(666)

    n_train = 25
    n_test = 20

    # List of dummy representations
    X = np.random.rand(n_train, 1000)
    Xs = np.random.rand(n_test, 1000)

    sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):
            Ktest[i, j] = np.exp(np.sum(np.square(X[i] - Xs[j])) / (-2.0 * sigma**2))

    K = gaussian_kernel(X, Xs, sigma)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in Gaussian kernel"

    Ksymm = gaussian_kernel(X, X, sigma)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in Gaussian kernel"

    Ksymm2 = gaussian_kernel_symmetric(X, sigma)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm2), "Error in Gaussian kernel"


def test_linear_kernel():

    np.random.seed(666)

    n_train = 25
    n_test = 20

    # List of dummy representations
    X = np.random.rand(n_train, 1000)
    Xs = np.random.rand(n_test, 1000)

    # UNUSED sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):
            Ktest[i, j] = np.dot(X[i], Xs[j])

    K = linear_kernel(X, Xs)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in linear kernel"

    Ksymm = linear_kernel(X, X)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in linear kernel"


def test_matern_kernel():

    np.random.seed(666)

    for metric in ("l1", "l2"):
        for order in (0, 1, 2):
            matern(metric, order)


def matern(metric, order):

    n_train = 25
    n_test = 20

    # List of dummy representations
    X = np.random.rand(n_train, 1000)
    Xs = np.random.rand(n_test, 1000)

    sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):

            if metric == "l1":
                d = np.sum(abs(X[i] - Xs[j]))
            else:
                d = np.sqrt(np.sum((X[i] - Xs[j]) ** 2))

            if order == 0:
                Ktest[i, j] = np.exp(-d / sigma)
            elif order == 1:
                Ktest[i, j] = np.exp(-np.sqrt(3) * d / sigma) * (1 + np.sqrt(3) * d / sigma)
            else:
                Ktest[i, j] = np.exp(-np.sqrt(5) * d / sigma) * (
                    1 + np.sqrt(5) * d / sigma + 5.0 / 3 * d**2 / sigma**2
                )

    K = matern_kernel(X, Xs, sigma, metric=metric, order=order)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in Matern kernel"

    Ksymm = matern_kernel(X, X, sigma, metric=metric, order=order)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in Matern kernel"


def test_sargan_kernel():

    np.random.seed(666)

    for ngamma in (0, 1, 2):
        sargan(ngamma)


def sargan(ngamma):

    n_train = 25
    n_test = 20

    gammas = np.random.random(ngamma)

    # List of dummy representations
    X = np.random.rand(n_train, 1000)
    Xs = np.random.rand(n_test, 1000)

    sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):
            d = np.sum(abs(X[i] - Xs[j]))

            factor = 1
            for k, gamma in enumerate(gammas):
                factor += gamma / sigma ** (k + 1) * d ** (k + 1)
            Ktest[i, j] = np.exp(-d / sigma) * factor

    K = sargan_kernel(X, Xs, sigma, gammas)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in Sargan kernel"

    Ksymm = sargan_kernel(X, X, sigma, gammas)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in Sargan kernel"


def array_nan_close(a, b):
    # Compares arrays, ignoring nans

    m = np.isfinite(a) & np.isfinite(b)
    return np.allclose(a[m], b[m], atol=1e-8, rtol=0.0)


def test_kpca():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenam
    data = get_energies(ASSETS / "hof_qm7.txt")

    keys = sorted(data.keys())

    np.random.seed(666)
    np.random.shuffle(keys)

    n_mols = 100

    representations = []

    for xyz_file in keys[:n_mols]:

        filename = ASSETS / "qm7" / xyz_file
        coordinates, atoms = read_xyz(filename)

        atomtypes = np.unique(atoms)
        representation = generate_bob(atoms, coordinates, atomtypes)
        representations.append(representation)

    X = np.array([representation for representation in representations])
    K = laplacian_kernel(X, X, 2e5)

    # calculate pca
    pcas_qml = kpca(K, n=10)

    # Calculate with sklearn
    pcas_sklearn = KernelPCA(10, eigen_solver="dense", kernel="precomputed").fit_transform(K)

    assert array_nan_close(
        np.abs(pcas_sklearn.T), np.abs(pcas_qml)
    ), "Error in Kernel PCA decomposition."


def test_wasserstein_kernel():

    np.random.seed(666)

    n_train = 5
    n_test = 3

    # List of dummy representations
    X = np.array(np.random.randint(0, 10, size=(n_train, 3)), dtype=np.float64)
    Xs = np.array(np.random.randint(0, 10, size=(n_test, 3)), dtype=np.float64)

    sigma = 100.0

    Ktest = np.zeros((n_train, n_test))

    for i in range(n_train):
        for j in range(n_test):
            Ktest[i, j] = np.exp(wasserstein_distance(X[i], Xs[j]) / (-1.0 * sigma))

    K = wasserstein_kernel(X, Xs, sigma)

    # Compare two implementations:
    assert np.allclose(K, Ktest), "Error in Wasserstein kernel"

    Ksymm = wasserstein_kernel(X, X, sigma)

    # Check for symmetry:
    assert np.allclose(Ksymm, Ksymm.T), "Error in Wasserstein kernel"



from setuptools import setup

try:
    import _compile
except ImportError:
    import sys
    from pathlib import Path

    sys.path.append(str(Path(__file__).resolve().parent))
    import _compile

if __name__ == "__main__":
    _compile.main()
    setup(
        description="Python/Fortran toolkit for representation of molecules and solids for machine learning of properties of molecules and solids.",
        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Topic :: Scientific/Engineering :: Chemistry",
        ],
        keywords=["qml", "quantum chemistry", "machine learning"],
    )



from typing import List, Optional

import numpy as np
from numpy import int64, ndarray

from .fslatm import fget_sbop, fget_sbop_local, fget_sbot, fget_sbot_local


def update_m(obj, ia, rcut=9.0, pbc=None):
    """
    retrieve local structure around atom ia
    for periodic systems (or very large system)
    """

    raise NotImplementedError("Function is in-complete")

    # zs, coords, c = obj
    # v1, v2, v3 = c
    # vs = spatial_distance.norm(c, axis=0)

    # nns = []
    # for i, vi in enumerate(vs):
    #     # n1_doulbe = rcut / li  # what is li
    #     # n1 = int(n1_doulbe)
    #     # if n1 - n1_doulbe == 0:
    #     #     n1s = (
    #     #         range(-n1, n1 + 1)
    #     #         if pbc[i]
    #     #         else [
    #     #             0,
    #     #         ]
    #     #     )
    #     # elif n1 == 0:
    #     #     n1s = (
    #     #         [-1, 0, 1]
    #     #         if pbc[i]
    #     #         else [
    #     #             0,
    #     #         ]
    #     #     )
    #     # else:
    #     #     n1s = (
    #     #         range(-n1 - 1, n1 + 2)
    #     #         if pbc[i]
    #     #         else [
    #     #             0,
    #     #         ]
    #     #     )

    #     # nns.append(n1s)

    # n1s, n2s, n3s = nns

    # n123s_ = np.array(list(itertools.product(n1s, n2s, n3s)))
    # n123s = []
    # for n123 in n123s_:
    #     n123u = list(n123)
    #     if n123u != [0, 0, 0]:
    #         n123s.append(n123u)

    # nau = len(n123s)
    # n123s = np.array(n123s, np.float64)

    # na = len(zs)
    # cia = coords[ia]

    # if na == 1:
    #     ds = np.array([[0.0]])
    # else:
    #     ds = spatial_distance.squareform(spatial_distance.pdist(coords))

    # zs_u = []
    # coords_u = []
    # zs_u.append(zs[ia])
    # coords_u.append(coords[ia])
    # for i in range(na):
    #     di = ds[i, ia]
    #     if (di > 0) and (di <= rcut):
    #         zs_u.append(zs[i])
    #         coords_u.append(coords[ia])

    #         # add new coords by translation
    #         ts = np.zeros((nau, 3))
    #         for iau in range(nau):
    #             ts[iau] = np.dot(n123s[iau], c)

    #         coords_iu = coords[i] + ts  # np.dot(n123s, c)
    #         dsi = spatial_distance.norm(coords_iu - cia, axis=1)
    #         filt = np.logical_and(dsi > 0, dsi <= rcut)
    #         nx = filt.sum()
    #         zs_u += [
    #             zs[i],
    #         ] * nx
    #         coords_u += [
    #             list(coords_iu[filt, :]),
    #         ]

    # obj_u = [zs_u, coords_u]

    # return obj_u


def get_boa(z1: int64, zs_: ndarray) -> ndarray:
    return z1 * np.array(
        [
            (zs_ == z1).sum(),
        ]
    )
    # return -0.5*z1**2.4*np.array( [(zs_ == z1).sum(), ])


def get_sbop(
    mbtype: List[int64],
    obj: List[ndarray],
    iloc: bool = False,
    ia: Optional[int] = None,
    normalize: bool = True,
    sigma: float = 0.05,
    rcut: float = 4.8,
    dgrid: float = 0.03,
    pbc: str = "000",
    rpower: int = 6,
) -> ndarray:
    """
    two-body terms

    :param obj: molecule object, consisting of two parts: [ zs, coords ]
    :type obj: list
    """

    z1, z2 = mbtype
    zs, coords, c = obj

    if iloc and ia is None:
        raise ValueError("Please specify za and ia")

    if pbc != "000":
        raise NotImplementedError("Periodic boundary conditions not implemented")
        # if rcut < 9.0:
        #     raise ValueError()
        # assert iloc, "#ERROR: for periodic system, plz use atomic rpst"
        # zs, coords = update_m(obj, ia, rcut=rcut, pbc=pbc)

        # after update of `m, the query atom `ia will become the first atom
        # ia = 0

    # bop potential distribution
    r0 = 0.1
    nx = int((rcut - r0) / dgrid) + 1

    coeff = 1 / np.sqrt(2 * sigma**2 * np.pi) if normalize else 1.0

    if iloc:
        ys = fget_sbop_local(coords, zs, ia, z1, z2, rcut, nx, dgrid, sigma, coeff, rpower)
    else:
        ys = fget_sbop(coords, zs, z1, z2, rcut, nx, dgrid, sigma, coeff, rpower)

    return ys


def get_sbot(
    mbtype: List[int64],
    obj: List[ndarray],
    iloc: bool = False,
    ia: Optional[int] = None,
    normalize: bool = True,
    sigma: float = 0.05,
    rcut: float = 4.8,
    dgrid: float = 0.0262,
    pbc: str = "000",
) -> ndarray:
    """
    sigma -- standard deviation of gaussian distribution centered on a specific angle
            defaults to 0.05 (rad), approximately 3 degree
    dgrid    -- step of angle grid
            defaults to 0.0262 (rad), approximately 1.5 degree
    """

    z1, z2, z3 = mbtype
    zs, coords, c = obj

    if iloc and ia is None:
        raise ValueError("Please specify za and ia")

    if pbc != "000":
        raise NotImplementedError("Periodic boundary conditions not implemented")
        # assert iloc, "#ERROR: for periodic system, plz use atomic rpst"
        # zs, coords = update_m(obj, ia, rcut=rcut, pbc=pbc)

        # # after update of `m, the query atom `ia will become the first atom
        # ia = 0

    # for a normalized gaussian distribution, u should multiply this coeff
    coeff = 1 / np.sqrt(2 * sigma**2 * np.pi) if normalize else 1.0

    # Setup grid in Python
    d2r = np.pi / 180  # degree to rad
    a0 = -20.0 * d2r
    a1 = np.pi + 20.0 * d2r
    nx = int((a1 - a0) / dgrid) + 1

    if iloc:
        ys = fget_sbot_local(coords, zs, ia, z1, z2, z3, rcut, nx, dgrid, sigma, coeff)
    else:
        ys = fget_sbot(coords, zs, z1, z2, z3, rcut, nx, dgrid, sigma, coeff)

    return ys



import numpy as np
from conftest import ASSETS, get_energies, shuffle_arrays

from qmllib.kernels import laplacian_kernel
from qmllib.representations import generate_coulomb_matrix
from qmllib.solvers import cho_solve
from qmllib.utils.xyz_format import read_xyz


def test_krr_cmat():

    # Parse file containing PBE0/def2-TZVP heats of formation and xyz filenames
    data = get_energies(ASSETS / "hof_qm7.txt")

    n_points = 1000

    all_representations = []
    all_properties = []

    filenames = sorted(data.keys())[:n_points]

    for filename in filenames:
        coord, atoms = read_xyz((ASSETS / "qm7" / filename).with_suffix(".xyz"))

        representation = generate_coulomb_matrix(atoms, coord, size=23, sorting="row-norm")

        all_representations.append(representation)
        all_properties.append(data[filename])

    all_representations = np.array(all_representations)
    all_properties = np.array(all_properties)

    shuffle_arrays(all_properties, all_representations, seed=666)

    # Make training and test sets
    n_test = 300
    n_train = 700
    train_indices = list(range(n_train))
    test_indices = list(range(n_train, n_train + n_test))

    # List of representations and properties
    test_representations = all_representations[test_indices]
    train_representations = all_representations[train_indices]
    test_properties = all_properties[test_indices]
    train_properties = all_properties[train_indices]

    # Set hyper-parameters
    sigma = 10 ** (4.2)
    llambda = 10 ** (-10.0)

    # Generate training Kernel
    K = laplacian_kernel(train_representations, train_representations, sigma)

    # Solve alpha
    K[np.diag_indices_from(K)] += llambda
    alpha = cho_solve(K, train_properties)

    # Calculate prediction kernel
    Ks = laplacian_kernel(train_representations, test_representations, sigma)
    predicted_properties = np.dot(Ks.transpose(), alpha)

    mae = np.mean(np.abs(test_properties - predicted_properties))

    assert mae < 6.0, "ERROR: Too high MAE!"



import ast
import csv
from copy import deepcopy

import numpy as np
from conftest import ASSETS

from qmllib.kernels import (
    get_atomic_local_gradient_kernel,
    get_atomic_local_kernel,
    get_gdml_kernel,
    get_global_kernel,
    get_gp_kernel,
    get_local_gradient_kernel,
    get_local_kernel,
    get_local_kernels,
    get_local_symmetric_kernel,
    get_local_symmetric_kernels,
    get_symmetric_gdml_kernel,
    get_symmetric_gp_kernel,
)
from qmllib.representations import generate_acsf

np.set_printoptions(linewidth=666, edgeitems=10)

CSV_FILE = ASSETS / "amons_small.csv"

TRAINING = 7
TEST = 5

SIGMA = 2.5
MAX_ATOMS = 20
DX = 1e-4


def csv_to_molecular_reps(csv_filename):

    np.random.seed(667)

    x = []
    dx = []
    f = []
    e = []
    n = []
    q = []

    disp_x = [[] for _ in range(4)]

    with open(csv_filename, "r") as csvfile:

        df = csv.reader(csvfile, delimiter=";", quotechar="#")

        for i, row in enumerate(df):

            if i > TEST + TRAINING:
                break

            coordinates = np.array(ast.literal_eval(row[2]))
            nuclear_charges = ast.literal_eval(row[5])
            # UNUSED atomtypes = ast.literal_eval(row[1])
            force = np.array(ast.literal_eval(row[3]))
            energy = float(row[6])

            (x1, dx1) = generate_acsf(nuclear_charges, coordinates, pad=MAX_ATOMS, gradients=True)

            x.append(x1)
            dx.append(dx1)
            f.append(force)
            e.append(energy)
            n.append(len(nuclear_charges))
            q.append(nuclear_charges)

            for j in range(len(nuclear_charges)):
                for xyz in range(3):

                    for k, disp in enumerate([2 * DX, DX, -DX, -2 * DX]):
                        disp_coords = deepcopy(coordinates)

                        disp_coords[j, xyz] += disp

                        # dx1 = generate_fchl_acsf(nuclear_charges, disp_coords, **REP_PARAMS)
                        # print(dx1.shape)
                        dx1 = generate_acsf(nuclear_charges, disp_coords, pad=MAX_ATOMS)

                        drep = dx1

                        disp_x[k].append(drep)

    return np.array(x), np.array(dx), e, f, np.array(disp_x), n, q


def test_global_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    Qs = Qall[-TEST:]

    K1 = get_global_kernel(X, Xs, Q, Qs, SIGMA)
    K2 = get_global_kernel(X, X, Q, Q, SIGMA)

    # TODO No assertion
    assert K1 is not None
    assert K2 is not None


def test_local_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    # UNUSED dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    # UNUSED dispX = dispXall[:, : sum(N) * 3, :, :]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    # UNUSED dXs = dXall[-TEST:]
    Ns = Nall[-TEST:]
    # UNUSED dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]
    Qs = Qall[-TEST:]

    K = get_local_kernel(X, X, Q, Q, SIGMA)
    assert np.allclose(K, K.T), "Error in symmetric get_local_kernel()"

    Ksymm = get_local_symmetric_kernel(X, Q, SIGMA)
    assert np.allclose(Ksymm, K), "Error in get_local_symmetric_kernel()"

    K = get_local_kernel(X, Xs, Q, Qs, SIGMA)

    K_numm = np.zeros(K.shape)

    for i in range(TRAINING):
        for n1 in range(N[i]):

            for j in range(TEST):
                for n2 in range(Ns[j]):
                    if Q[i][n1] == Qs[j][n2]:

                        d = np.linalg.norm(X[i, n1] - Xs[j, n2])
                        gauss = np.exp(-(d**2) / (2 * SIGMA**2))
                        K_numm[j, i] += gauss

    assert np.allclose(K, K_numm), "Error in get_local_kernel()"


def test_atomic_local_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    # UNUSED dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    # UNUSED dispX = dispXall[:, : sum(N) * 3, :, :]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    # UNUSED dXs = dXall[-TEST:]
    Ns = Nall[-TEST:]
    # UNUSED dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]
    Qs = Qall[-TEST:]

    K = get_atomic_local_kernel(X, Xs, Q, Qs, SIGMA)

    K_numm = np.zeros(K.shape)

    idx = 0

    for i in range(TRAINING):
        for n1 in range(N[i]):

            for j in range(TEST):
                for n2 in range(Ns[j]):

                    if Q[i][n1] == Qs[j][n2]:
                        d = np.linalg.norm(X[i, n1] - Xs[j, n2])
                        gauss = np.exp(-(d**2) / (2 * SIGMA**2))
                        K_numm[j, idx] += gauss

            idx += 1

    assert np.allclose(K, K_numm), "Error in get_local_kernel()"


def test_atomic_local_gradient():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]  # noqa:F841
    dXs = dXall[-TEST:]  # noqa:F841
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]  # noqa:F841
    Qs = Qall[-TEST:]  # noqa:F841

    Kt_gradient = get_atomic_local_gradient_kernel(X, X, dX, Q, Q, SIGMA)

    idx1 = 0

    K_numm = np.zeros((4, Kt_gradient.shape[0], Kt_gradient.shape[1]))

    for i in range(TRAINING):
        for n1 in range(N[i]):

            idx2 = 0
            for j in range(TRAINING):
                for n2 in range(N[j]):
                    for xyz in range(3):

                        for n_diff in range(N[j]):
                            for k in range(4):
                                if Q[i][n1] == Q[j][n_diff]:
                                    d = np.linalg.norm(X[i, n1] - dispX[k, idx2, n_diff])
                                    gauss = np.exp(-(d**2) / (2 * SIGMA**2))
                                    K_numm[k, idx2, idx1] += gauss
                        idx2 += 1
            idx1 += 1

    K_numm = -(-K_numm[0] + 8 * K_numm[1] - 8 * K_numm[2] + K_numm[3]) / (12 * DX)

    assert np.allclose(Kt_gradient, K_numm), "Error in get_atomic_local_gradient_kernel()"


def test_local_gradient():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]  # noqa:F841
    dXs = dXall[-TEST:]  # noqa:F841
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]  # noqa:F841
    Qs = Qall[-TEST:]  # noqa:F841

    Kt_gradient = get_local_gradient_kernel(X, X, dX, Q, Q, SIGMA)

    idx2 = 0

    K_numm = np.zeros((4, Kt_gradient.shape[0], Kt_gradient.shape[1]))

    for i in range(TRAINING):
        for n1 in range(N[i]):

            idx1 = 0

            for j in range(TRAINING):
                for n2 in range(N[j]):
                    for xyz in range(3):
                        for n_diff in range(N[j]):
                            for k in range(4):

                                if Q[i][n1] == Q[j][n_diff]:
                                    d = np.linalg.norm(X[i, n1] - dispX[k, idx1, n_diff])
                                    gauss = np.exp(-(d**2) / (2 * SIGMA**2))
                                    K_numm[k, idx1, idx2] += gauss

                        idx1 += 1
        idx2 += 1

    K_numm = -(-K_numm[0] + 8 * K_numm[1] - 8 * K_numm[2] + K_numm[3]) / (12 * DX)

    assert np.allclose(Kt_gradient, K_numm), "Error in get_local_gradient_kernel()"


def test_gdml_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]
    Qs = Qall[-TEST:]

    Kt_gdml = get_gdml_kernel(X, Xs, dX, dXs, Q, Qs, SIGMA)

    idx1 = 0

    K_numm = np.zeros(Kt_gdml.shape)

    for i in range(TRAINING):
        for n1 in range(N[i]):
            for xyz1 in range(3):
                for n_diff1 in range(N[i]):

                    idx2 = 0

                    for j in range(TEST):
                        for n2 in range(Ns[j]):
                            for xyz2 in range(3):
                                for n_diff2 in range(Ns[j]):

                                    if Q[i][n_diff1] == Qs[j][n_diff2]:
                                        # displacements = [2*DX, DX, -DX, -2*DX]

                                        d = np.linalg.norm(
                                            dispX[1, idx1, n_diff1] - dispXs[1, idx2, n_diff2]
                                        )
                                        gauss1 = np.exp(-(d**2) / (2 * SIGMA**2))

                                        d = np.linalg.norm(
                                            dispX[1, idx1, n_diff1] - dispXs[2, idx2, n_diff2]
                                        )
                                        gauss2 = np.exp(-(d**2) / (2 * SIGMA**2))

                                        d = np.linalg.norm(
                                            dispX[2, idx1, n_diff1] - dispXs[1, idx2, n_diff2]
                                        )
                                        gauss3 = np.exp(-(d**2) / (2 * SIGMA**2))

                                        d = np.linalg.norm(
                                            dispX[2, idx1, n_diff1] - dispXs[2, idx2, n_diff2]
                                        )
                                        gauss4 = np.exp(-(d**2) / (2 * SIGMA**2))

                                        gauss = (gauss1 - gauss2 - gauss3 + gauss4) / (4 * DX * DX)
                                        K_numm[idx2, idx1] += gauss

                                idx2 += 1

                idx1 += 1

    # Numerical hessian has quite large error, probably around 1e-7, so setting atol to 1e-6.
    assert np.allclose(Kt_gdml, K_numm, atol=1e-6), "Error in get_gdml_kernel()"


def test_symmetric_gdml_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]  # noqa:F841
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]  # noqa:F841
    dXs = dXall[-TEST:]  # noqa:F841
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]  # noqa:F841
    Qs = Qall[-TEST:]  # noqa:F841

    K = get_gdml_kernel(X, X, dX, dX, Q, Q, SIGMA)
    K_symm = get_symmetric_gdml_kernel(X, dX, Q, SIGMA)

    assert np.allclose(K, K_symm), "Error in get_symmetric_gdml_kernel()"


def test_gp_kernel():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]  # noqa:F841
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]  # noqa:F841
    Qs = Qall[-TEST:]

    K = get_gp_kernel(X, Xs, dX, dXs, Q, Qs, SIGMA)

    K_uu = get_local_kernel(X, Xs, Q, Qs, SIGMA)
    assert np.allclose(K[:TEST, :TRAINING], K_uu), "Error: Fail in Gaussian Process kernel K_uu"

    K_ug = get_local_gradient_kernel(X, Xs, dXs, Q, Qs, SIGMA)
    assert np.allclose(K[TEST:, :TRAINING], K_ug), "Error: Fail in Gaussian Process kernel K_ug"

    K_gu = get_local_gradient_kernel(Xs, X, dX, Qs, Q, SIGMA)
    assert np.allclose(K[:TEST, TRAINING:], K_gu.T), "Error: Fail in Gaussian Process kernel K_gu"

    K_gg = get_gdml_kernel(X, Xs, dX, dXs, Q, Qs, SIGMA)

    assert np.allclose(K[TEST:, TRAINING:], K_gg), "Error: Fail in Gaussian Process kernel K_gg"

    K1_symm = get_gp_kernel(X, X, dX, dX, Q, Q, SIGMA)
    K2_symm = get_symmetric_gp_kernel(X, dX, Q, SIGMA)
    assert np.allclose(K1_symm, K2_symm), "Error: Fail in Gaussian Process symmetric kernel"


def test_local_kernels():

    Xall, dXall, Eall, Fall, dispXall, Nall, Qall = csv_to_molecular_reps(CSV_FILE)

    X = Xall[:TRAINING]
    dX = dXall[:TRAINING]  # noqa:F841
    N = Nall[:TRAINING]
    dispX = dispXall[:, : sum(N) * 3, :, :]  # noqa:F841
    Q = Qall[:TRAINING]

    Xs = Xall[-TEST:]
    dXs = dXall[-TEST:]  # noqa:F841
    Ns = Nall[-TEST:]
    dispXs = dispXall[:, -sum(Ns) * 3 :, :, :]  # noqa:F841
    Qs = Qall[-TEST:]

    sigma1 = 2.5
    sigma2 = 25.0

    K1 = get_local_kernel(X, Xs, Q, Qs, sigma1)
    K2 = get_local_kernel(X, Xs, Q, Qs, sigma2)
    K3 = get_local_kernel(X, Xs, Q, Qs, 10.0)  # noqa:F841
    K4 = get_local_kernel(X, Xs, Q, Qs, 2.0)  # noqa:F841

    K5 = get_local_kernel(X, Xs, Q, Qs, 3.0)  # noqa:F841

    K = get_local_kernels(X, Xs, Q, Qs, [sigma1, sigma2, 10.0, 2.0, 3.0])

    assert np.allclose(K1, K[0]), "Error in get_local_kernels() 1"
    assert np.allclose(K2, K[1]), "Error in get_local_kernels() 2"

    K1 = get_local_kernel(X, X, Q, Q, sigma1)
    K2 = get_local_symmetric_kernel(X, Q, sigma2)

    K = get_local_symmetric_kernels(X, Q, [sigma1, sigma2])

    assert np.allclose(K1, K[0]), "Error in get_local_symmetric_kernels() 1"
    assert np.allclose(K2, K[1]), "Error in get_local_symmetric_kernels() 2"
